'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _kindOf = require('kind-of');

var _kindOf2 = _interopRequireDefault(_kindOf);

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _types = require('./types');

var _types2 = _interopRequireDefault(_types);

var _schemas = require('./schemas');

var _schemas2 = _interopRequireDefault(_schemas);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * A private string to identify structs by.
 *
 * @type {String}
 */

const IS_STRUCT = '@@__STRUCT__@@';

/**
 * Public properties for struct functions.
 *
 * @type {Array}
 */

const STRUCT_PROPERTIES = ['kind', 'type', 'schema', 'defaults', 'options'];

/**
 * Public methods for struct functions.
 *
 * @type {Array}
 */

const STRUCT_METHODS = ['assert', 'default', 'test', 'validate'];

/**
 * Create a struct factory with a `config`.
 *
 * @param {Object} config
 * @return {Function}
 */

function superstruct(config = {}) {
  const types = _extends({}, _types2.default, config.types || {});

  /**
   * Create a `kind` struct with schema `definition`, `defaults` and `options`.
   *
   * @param {String} kind
   * @param {Function|String|Array|Object} definition
   * @param {Any} defaults
   * @param {Object} options
   * @return {Function}
   */

  function createStruct(kind, definition, defaults, options = {}) {
    if (isStruct(definition)) {
      definition = definition.schema;
    }

    const Schema = _schemas2.default[kind];
    const schema = new Schema(definition, defaults, _extends({}, options, { types, struct }));

    // Define the struct creator function.
    function Struct(data) {
      !!(this instanceof Struct) ? process.env.NODE_ENV !== 'production' ? (0, _invariant2.default)(false, 'The `Struct` creation function should not be used with the `new` keyword.') : (0, _invariant2.default)(false) : void 0;
      return schema.assert(data);
    }

    Struct[IS_STRUCT] = true;
    Struct.kind = kind;

    STRUCT_PROPERTIES.forEach(prop => {
      Struct[prop] = schema[prop];
    });

    STRUCT_METHODS.forEach(method => {
      Struct[method] = (...a) => schema[method](...a);
    });

    return Struct;
  }

  /**
   * Define a struct with schema `definition`, `defaults` and `options`.
   *
   * @param {Function|String|Array|Object} definition
   * @param {Any} defaults
   * @param {Object} options
   * @return {Function}
   */

  function struct(definition, defaults, options) {
    if (isStruct(definition)) return definition;
    const kind = getKind(definition);
    const Struct = createStruct(kind, definition, defaults, options);
    return Struct;
  }

  // Mix in a factory for each kind of struct.
  Object.keys(_schemas2.default).forEach(kind => {
    const lower = kind.toLowerCase();
    struct[lower] = (...args) => createStruct(kind, ...args);
  });

  // Mix in the `required` convenience flag.
  Object.defineProperty(struct, 'required', {
    get: () => (s, d, o = {}) => struct(s, d, _extends({}, o, { required: true }))
  });

  /**
   * Return the struct factory.
   */

  return struct;
}

/**
 * Check if a `value` is a struct.
 *
 * @param {Any} value
 * @return {Boolean}
 */

function isStruct(value) {
  return !!(value && value[IS_STRUCT]);
}

/**
 * Get the kind of a struct from its schema `definition`.
 *
 * @param {Any} definition
 * @return {String}
 */

function getKind(definition) {
  switch ((0, _kindOf2.default)(definition)) {
    case 'function':
      return 'Function';
    case 'string':
      return 'Scalar';
    case 'array':
      return 'List';
    case 'object':
      return 'Object';
    default:
      {
        !false ? process.env.NODE_ENV !== 'production' ? (0, _invariant2.default)(false, `A struct schema definition must be a string, array or object, but you passed: ${definition}`) : (0, _invariant2.default)(false) : void 0;
      }
  }
}

/**
 * Export.
 *
 * @type {Function}
 */

exports.default = superstruct;